{{
'   Started as fsrw.spin 1.4  4 February 2007   Radical Eye Software
'
'   Modified by Jim C in order to read, format and output the boot sector of
'   an SD card that has been formated FAT16
'
'   For additional details, refer to fsrw.spin
'
'   Outputs to terminal program on PC, but that can easily be changed
'  }}
'
'   Constants describing FAT volumes.
'
con
        _clkmode        = xtal1 + pll16x
        _xinfreq        = 5_000_000
   SECTORSIZE = 512
   SECTORSHIFT = 9
   DIRSIZE = 32
   DIRSHIFT = 5
'
'   The object that provides the block-level access.
'
obj
   sdspi: "sdspiqasm"
   debug: "PC_DebugDemo"
var
'
'
'   Variables concerning the open file.
'
   long fclust ' the current cluster number
   long filesize ' the total current size of the file
   long floc ' the seek position of the file
   long frem ' how many bytes remain in this cluster from this file
   long bufat ' where in the buffer our current character is
   long bufend ' the last valid character (read) or free position (write)
   long direntry ' the byte address of the directory entry (if open for write)
   long writelink ' the byte offset of the disk location to store a new cluster
   long fatptr ' the byte address of the most recently written fat entry
'
'   Variables used when mounting to describe the FAT layout of the card.
'
   long rootdir ' the byte address of the start of the root directory
   long rootdirend ' the byte immediately following the root directory.
   long dataregion ' the start of the data region, offset by two sectors
   long clustershift ' log base 2 of blocks per cluster
   long fat1 ' the block address of the fat1 space
   long totclusters ' how many clusters in the volume
   long sectorsperfat ' how many sectors per fat
'
'   Variables controlling the caching.
'
   long lastread ' the block address of the buf2 contents
   long dirty ' nonzero if buf2 is dirty
   long t1
'
'  Buffering:  two sector buffers.  These two buffers must be longword
'  aligned!  To ensure this, make sure they are the first byte variables
'  defined in this object.
'
   byte buf[SECTORSIZE] '  main data buffer
   byte buf2[SECTORSIZE] ' main metadata buffer
   byte padname[11] ' filename buffer
   byte tbuf[20],byter[4]
pub StartSD | r, sta, bytes
   debug.start(230400)
   mount(0)
   debug.str(string("SD Card Mounted", 13))
   debug.str(string("Dir: ", 13))
   opendir
   repeat while 0 == nextfile(@tbuf)
      debug.str(@tbuf)
      debug.out(13)
   debug.str(string("That's the dir", 13))
   debug.out(13)
   r := popen(string("newfilex.txt"), "w")
   debug.str(string("Opening returned "))
   debug.dec(r)
   debug.out(13)
   repeat


pub opendir | off
{{
'   Close the currently open file, and set up the read buffer for
'   calls to nextfile.
}}
   pclose
   off := rootdir - (dataregion << SECTORSHIFT)
   fclust := off >> (clustershift + SECTORSHIFT)
   floc := off - (fclust << (clustershift + SECTORSHIFT))
   frem := rootdirend - rootdir
   filesize := floc + frem
   return 0

pub nextfile(fbuf) | i, t, at, lns
{{
'   Find the next file in the root directory and extract its
'   (8.3) name into fbuf.  Fbuf must be sized to hold at least
'   13 characters (8 + 1 + 3 + 1).  If there is no next file,
'   -1 will be returned.  If there is, 0 will be returned.
}}
   repeat
      if (bufat => bufend)
         t := pfillbuf
         if (t < 0)
            return t
      at := @buf + bufat
      if (byte[at] == 0)
         return -1
      bufat += DIRSIZE
      if (byte[at] <> $e5 and (byte[at][$0b] & $18) == 0)
         lns := fbuf
         repeat i from 0 to 10
            byte[fbuf] := byte[at][i]
            fbuf++
            if (byte[at][i] <> " ")
               lns := fbuf
            if (i == 7 or i == 10)
               fbuf := lns
               if (i == 7)
                  byte[fbuf] := "."
                  fbuf++
         byte[fbuf] := 0
         return 0

pub popen(s, mode) | i, sentinel, dirptr, freeentry
{{
'   Close any currently open file, and open a new one with the given
'   file name and mode.  Mode can be "r" "w" "a" or "d" (delete).
'   If the file is opened successfully, 0 will be returned.  If the
'   file did not exist, and the mode was not "w" or "a", -1 will be
'   returned.  Otherwise abort will be called with a negative error
'   code.
}}
   pclose
   i := 0
   repeat while (i<8 and byte[s] and byte[s] <> ".")
      padname[i++] := uc(byte[s++])
   repeat while (i<8)
      padname[i++] := " "
   repeat while (byte[s] and byte[s] <> ".")
      s++
   if (byte[s] == ".")
      s++
   repeat while (i<11 and byte[s])
      padname[i++] := uc(byte[s++])
   repeat while (i < 11)
      padname[i++] := " "
   sentinel := 0
   freeentry := 0
   repeat dirptr from rootdir to rootdirend - DIRSIZE step DIRSIZE
      s := readbytec(dirptr)
      if (freeentry == 0 and (byte[s] == 0 or byte[s] == $e5))
         freeentry := dirptr
      if (byte[s] == 0)
         sentinel := dirptr
         quit
      repeat i from 0 to 10
         if (padname[i] <> byte[s][i])
            quit
      if (i == 11 and 0 == (byte[s][$0b] & $18)) ' this always returns
         fclust := brword(s+$1a)
         filesize := brlong(s+$1c)
         if (mode == "r")
            frem := SECTORSIZE << clustershift
            if (frem > filesize)
               frem := filesize
            return 0
         if (byte[s][11] & $d9)
            abort(-6) ' no permission to write
         if (mode == "d")
            brwword(s, $e5)
            freeclusters(fclust)
            flushifdirty
            return 0
         if (mode == "w")
            brwword(s+26, -1)
            brwlong(s+28, 0)
            writelink := dirptr + 26
            direntry := dirptr
            freeclusters(fclust)
            bufend := SECTORSIZE
            fclust := 0
            filesize := 0
            frem := 0
            return 0
         elseif (mode == "a")
' this code will eventually be moved to seek
            frem := filesize
            freeentry := SECTORSIZE << clustershift
            if (fclust => $fff0)
               fclust := 0
            repeat while (frem > freeentry)
               if (fclust < 2)
                  abort(-7) ' eof repeat while following chain
               fclust := nextcluster
               frem -= freeentry
            floc := filesize & constant(!(SECTORSIZE - 1))
            bufend := SECTORSIZE
            bufat := frem & constant(SECTORSIZE - 1)
            writelink := dirptr + 26
            direntry := dirptr
            if (bufat)
               sdspi.readblock(datablock, @buf)
               frem := freeentry - (floc & (freeentry - 1))
            else
               if (fclust < 2 or frem == freeentry)
                  frem := 0
               else
                  frem := freeentry - (floc & (freeentry - 1))
            if (fclust => 2)
               followchain
            return 0
         else
            abort(-3) ' bad argument
   if (mode <> "w" and mode <> "a")
      return -1 ' not found
   direntry := freeentry
   if (direntry == 0)
      abort(-2) ' no empty directory entry
   ' write (or new append): create valid directory entry
   s := readbytec(direntry)
   bytefill(s, 0, DIRSIZE)
   bytemove(s, @padname, 11)
   brwword(s+26, -1)
   i := pdate
   brwlong(s+$e, i) ' write create time and date
   brwlong(s+$16, i) ' write last modified date and time
   if (direntry == sentinel and direntry + DIRSIZE < rootdirend)
      brwword(readbytec(direntry+DIRSIZE), 0)
   flushifdirty
   writelink := direntry + 26
   fclust := 0
   bufend := SECTORSIZE
   return 0
pub pread(ubuf, count) | r, t
{{
'   Read count bytes into the buffer ubuf.  Returns the number of bytes
'   successfully read, or a negative number if there is an error.
'   The buffer may be as large as you want.
}}
   r := 0
   repeat while (count > 0)
      if (bufat => bufend)
         t := pfillbuf
         if (t =< 0)
            if (r > 0)
               return r
            return t
      t := bufend - bufat
      if (t > count)
         t := count
      bytemove(ubuf, @buf+bufat, t)
      bufat += t
      r += t
      ubuf += t
      count -= t
   return r
pub pgetc | t
{{
'   Read and return a single character.  If the end of file is
'   reached, -1 will be returned.  If an error occurs, a negative
'   number will be returned.
}}
   if (bufat => bufend)
      t := pfillbuf
      if (t =< 0)
         return -1
   return (buf[bufat++])
pub pwrite(ubuf, count) | r, t
{{
'   Write count bytes from the buffer ubuf.  Returns the number of bytes
'   successfully written, or a negative number if there is an error.
'   The buffer may be as large as you want. 
'   (note: work through with ubuf = 2 x 512, and it's filled somewhere else)
'   (bufat starts at 0, and for write bufend starts at SECTORSIZE - 1
}}
   t := 0
   repeat while (count > 0)
      if (bufat => bufend)         ' 1st: bufat = 0, bufend = 511 2nd: bufat = 512 > bufend = 511 > buf flushed 
         t := pflushbuf(bufat, 0)  ' does pflushbuf modify bufat? Yes, sometimes, to 0 
      t := bufend - bufat          ' 1st: t = bufend.  2nd: bufat points to start of second sector ful
      if (t > count)               ' first time, t < count;  2nd: t = count
         t := count
      bytemove(@buf+bufat, ubuf, t)   'move a whole sector-ful from ubuf to buf[0]
      r += t                       ' r now t+1 (start of next sector-ful).  Is r used here??
      bufat += t                   ' bufat = 513 (or 512)
      ubuf += t                    ' same with ubuf pointer
      count -= t                   ' subtract a sector-ful from count Count = 512 after first pass
   return t
pub pputc(c)
{{
'   Write a single character into the file open for write.  Returns
'   0 if successful, or a negative number if some error occurred.
}}
   buf[bufat++] := c
   if (bufat == SECTORSIZE)
      return pflushbuf(SECTORSIZE, 0)
   return 0

pub pclose | r
{{
'   Flush and close the currently open file if any.  Also reset the
'   pointers to valid values.  If there is no error, 0 will be returned.
}}
   r := 0
   if (direntry)
      r := pflush
   bufat := 0
   bufend := 0
   filesize := 0
   floc := 0
   frem := 0
   writelink := 0
   direntry := 0
   fclust := 0
   return r

pub pflush
{{
'   Call flush with the current data buffer location, and the flush
'   metadata flag set.
}}
   return pflushbuf(bufat, 1)

pub mount(basepin) | start, sectorspercluster, reserved, rootentries, sectors
{{
'   Mount a volume.  The address passed in is passed along to the block
'   layer; see the currently used block layer for documentation.  If the
'   volume mounts, a 0 is returned, else abort is called.
}}
   sdspi.start(basepin)
   lastread := -1
   dirty := 0
   sdspi.readblock(0, @buf)         'this reads the first 512 bytes on volume: Bootsector   
   if (brlong(@buf+$36) == constant("F" + ("A" << 8) + ("T" << 16) + ("1" << 24)))
      start := 0
   else
      start := brlong(@buf+$1c6)
      sdspi.readblock(start, @buf)
   if (brlong(@buf+$36) <> constant("F" + ("A" << 8) + ("T" << 16) + ("1" << 24)) or buf[$3a] <> "6")
      abort(-20) ' not a fat16 volume
   if (brword(@buf+$0b) <> SECTORSIZE)
      abort(-21) ' bad bytes per sector
   sectorspercluster := buf[$0d]
   if (sectorspercluster & (sectorspercluster - 1))
      abort(-22) ' bad sectors per cluster
   clustershift := 0
   repeat while (sectorspercluster > 1)
      clustershift++
      sectorspercluster >>= 1

'   debug.str(string("Clustershift         : "))
'   debug.dec(clustershift)
'   debug.out(13)

   sectorspercluster := 1 << clustershift
   reserved := brword(@buf+$0e)
   if (buf[$10] <> 2)
      abort(-23) ' not two FATs
   rootentries := brword(@buf+$11)
   sectors := brword(@buf+$13)
   sectorsperfat := brword(@buf+$16)

'*********************** Boot Sector Data **************************   
   debug.str(string("Jump Instr:         "))
   debug.str(string(" "))
   debug.dec(buf[$00])
   debug.str(string(" "))
   debug.dec(buf[$01])
   debug.str(string(" "))
   debug.dec(buf[$02])
   debug.out(13)
   
   debug.str(string("OEM Names:           "))
   debug.out(buf[$03])
   debug.out(buf[$04])
   debug.out(buf[$05])
   debug.out(buf[$06])
   debug.out(buf[$07])
   debug.out(buf[$08])
   debug.out(buf[$09])
   debug.out(buf[$0a])
   debug.out(13)


   debug.str(string("Bytes per Sector:    "))
   debug.dec(brword(@buf+$0b))
   debug.out(13)

   debug.str(string("Sectors per Cluster: "))
   debug.dec(buf[$0d])
   debug.out(13)

   debug.str(string("Reserved Sector Cnt: "))
   t1 := brword(@buf+$0e)
   debug.dec(t1)
   debug.out(13)

   debug.str(string("Number of FAT's:     "))
   debug.dec(buf[$10])
   debug.out(13)

   debug.str(string("Max Directry Entries "))
   t1 := brword(@buf+$11)
   debug.dec(t1)
   debug.out(13)

   if (sectors == 0)
      sectors := brlong(@buf+$20)
   debug.str(string("Total Sectors        "))
   debug.dec(sectors)
   debug.out(13)

   debug.str(string("Media Descr.:        "))
   debug.ihex(buf[$15],2)
   debug.out(13)

   debug.str(string("Sectors per FAT:     "))
   t1 := brword(@buf+$16)
   debug.dec(t1)
   debug.out(13)

   debug.str(string("Sectors per Track:   "))
   t1 := brword(@buf+$18)
   debug.dec(t1)
   debug.out(13)

   debug.str(string("Number of Heads:     "))
   t1 := brword(@buf+$1a)
   debug.dec(t1)
   debug.out(13)

   debug.str(string("Hidden Sectors:      "))
   t1 := brlong(@buf+$1c)
   debug.dec(t1)
   debug.out(13)

   debug.str(string("Physical Driver Num: "))
   debug.ihex(buf[$24],2)
   debug.out(13)

   debug.str(string("Reserved:            "))
   debug.ihex(buf[$25],2)
   debug.out(13)

   debug.str(string("Signature:           "))
   debug.ihex(buf[$26],2)
   debug.out(13)

   debug.str(string("ID (Serial Number:   "))
   t1 := brlong(@buf+$27)
   debug.ihex(t1,8)
   debug.out(13)

   debug.str(string("Volume Label:        "))
   debug.out(buf[$2b])
   debug.out(buf[$2c])
   debug.out(buf[$2d])
   debug.out(buf[$2e])
   debug.out(buf[$2f])
   debug.out(buf[$30])
   debug.out(buf[$31])
   debug.out(buf[$32])
   debug.out(buf[$33])
   debug.out(buf[$34])
   debug.out(buf[$35])
   debug.out(13)

   debug.str(string("FAT File Type:       "))
   debug.out(buf[$36])
   debug.out(buf[$37])
   debug.out(buf[$38])
   debug.out(buf[$39])
   debug.out(buf[$3a])
   debug.out(buf[$3b])
   debug.out(13)

   debug.str(string("End of Sector Marker "))
   debug.ihex((brword(@buf+$1fe)),4)
   debug.out(13)

   if (brword(@buf+$1fe) <> $aa55)   
      abort(-24) ' bad FAT signature
   fat1 := start + reserved
   rootdir := (fat1 + 2 * sectorsperfat) << SECTORSHIFT
   rootdirend := rootdir + (rootentries << DIRSHIFT)
   dataregion := 1 + ((rootdirend - 1) >> SECTORSHIFT) - 2 * sectorspercluster
   totclusters := ((sectors - dataregion + start) >> clustershift)
   if (totclusters > $fff0)
      abort(-25) ' too many clusters
   return 0
pri writeblock2(n, b)
'
'   On metadata writes, if we are updating the FAT region, also update
'   the second FAT region.
'
   sdspi.writeblock(n, b)
   if (n => fat1 and n < fat1 + sectorsperfat)
      sdspi.writeblock(n+sectorsperfat, b)
pri flushifdirty
'
'   If the metadata block is dirty, write it out.
'
   if (dirty)
      writeblock2(lastread, @buf2)
      dirty := 0
pri readblockc(n)
'
'   Read a block into the metadata buffer, if that block is not already
'   there.
'
   if (n <> lastread)
      flushifdirty
      sdspi.readblock(n, @buf2)
      lastread := n
pri brword(b)
'
'   Read a byte-reversed word from a (possibly odd) address.
'
   return (byte[b]) + ((byte[b][1]) << 8)
pri brlong(b)
'
'   Read a byte-reversed long from a (possibly odd) address.
'
   return brword(b) + (brword(b+2) << 16)
pri brwword(w, v)
'
'   Write a byte-reversed word to a (possibly odd) address, and
'   mark the metadata buffer as dirty.
'
   byte[w++] := v
   byte[w] := v >> 8
   dirty := 1
pri brwlong(w, v)
'
'   Write a byte-reversed long to a (possibly odd) address, and
'   mark the metadata buffer as dirty.
'
   brwword(w, v)
   brwword(w+2, v >> 16)
pri readbytec(byteloc)
'
'   Read a byte address from the disk through the metadata buffer and
'   return a pointer to that location.
'
   readblockc(byteloc >> SECTORSHIFT)
   return @buf2 + (byteloc & constant(SECTORSIZE - 1))
pri readfat(clust)
'
'   Read a fat location and return a pointer to the location of that
'   entry.
'
   fatptr := (fat1 << SECTORSHIFT) + (clust << 1)
   return readbytec(fatptr)
pri followchain | clust
'
'   Follow the fat chain and update the writelink.
'
   clust := brword(readfat(fclust))
   writelink := fatptr
   return clust
pri nextcluster | clust
'
'   Read the next cluster and return it.  Set up writelink to
'   point to the cluster we just read, for later updating.  If the
'   cluster number is bad, return a negative number.
'
   clust := followchain
   if (clust < 2 or clust => totclusters)
      abort(-9) ' bad cluster value
   return clust
pri freeclusters(clust) | bp
'
'   Free an entire cluster chain.  Used by remove and by overwrite.
'   Assumes the pointer has already been cleared/set to $ffff.
'
   repeat while (clust < $fff0)
      if (clust < 2)
         abort(-26) ' bad cluster number")
      bp := readfat(clust)
      clust := brword(bp)
      brwword(bp, 0)
   flushifdirty
pri datablock
'
'   Calculate the block address of the current data location.
'
   return (fclust << clustershift) + dataregion + ((floc >> SECTORSHIFT) & ((1 << clustershift) - 1))
pri uc(c)
'
'   Compute the upper case version of a character.
'
   if ("a" =< c and c =< "z")
      return c - 32
   return c
pri pflushbuf(r, metadata) | cluststart, newcluster, count, i
'   start w/ r = bufat = 0   frem = 0 for write
'   Flush the current buffer, if we are open for write.  This may
'   allocate a new cluster if needed.  If metadata is true, the
'   metadata is written to disk, including any FAT cluster allocations
'   and the file size in the directory entry.
'
   if (direntry == 0)
      abort(-27) ' not open for writing
   if (r > 0) ' must *not* allocate cluster if flushing an empty buffer
      if (frem < SECTORSIZE)
         ' find a new cluster. Could be anywhere!  If possible, stay on the
         ' same page used for the last cluster.
         newcluster := -1
         cluststart := fclust & constant(!((SECTORSIZE >> 1) - 1))
         count := 2
         repeat
            readfat(cluststart)
            repeat i from 0 to constant(SECTORSIZE - 2) step 2
               if (buf2[i]==0 and buf2[i+1]==0)
                  quit
            if (i < SECTORSIZE)
               newcluster := cluststart + (i >> 1)
               if (newcluster => totclusters)
                  newcluster := -1
            if (newcluster > 1)
               brwword(@buf2+i, -1)
               brwword(readbytec(writelink), newcluster)
               writelink := fatptr + i
               fclust := newcluster
               frem := SECTORSIZE << clustershift
               quit
            else
               cluststart += constant(SECTORSIZE >> 1)
               if (cluststart => totclusters)
                  cluststart := 0
                  count--
                  if (count < 0)
                     r := -5 ' No space left on device
                     quit
      if (frem => SECTORSIZE)
         sdspi.writeblock(datablock, @buf)
         if (r == SECTORSIZE) ' full buffer, clear it
            floc += r
            frem -= r
            bufat := 0
            bufend := r
         else
            ' not a full blockleave pointers alone
   if (r < 0 or metadata) ' update metadata even if error
      readblockc(direntry >> SECTORSHIFT) ' flushes unwritten FAT too
      brwlong(@buf2+(direntry & constant(SECTORSIZE-1))+28, floc+bufat)
      flushifdirty
   if (r < 0)
      abort(r)
   return r
pri pfillbuf | r
'
'   Get some data into an empty buffer.  If no more data is available,
'   return -1.  Otherwise return the number of bytes read into the
'   buffer.
'
   if (floc => filesize)
      return -1
   if (frem == 0)
      fclust := nextcluster
      frem := SECTORSIZE << clustershift
      if (frem + floc > filesize)
         frem := filesize - floc
   sdspi.readblock(datablock, @buf)
   r := SECTORSIZE
   if (floc + r => filesize)
      r := filesize - floc
   floc += r
   frem -= r
   bufat := 0
   bufend := r
   return r
pri pdate
{{
'   Get the current date and time, as a long, in the format required
'   by FAT16.  Right now it"s hardwired to return the date this
'   software was created on (December 26, 2006).  You can change this
'   to return a valid date/time if you have access to this data in
'   your setup.
}}
   return constant(((2007-1980) << 25) + (1 << 21) + (6 << 16) + (12 << 11))